# Copyright 2019 Cumulus Networks, Inc.  All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

from netq_lib.common.netq_pq import NetQPriorityQueue, QueueElem
from collections import defaultdict
import enum
import time


class FullState(object):

    def __init__(self, msg_type):
        self.decorated = self
        self.msg_type = msg_type
        self.fullstate_proc_delay = 1

    def __call__(self, func):
        def wrapper(obj):
            if obj.full_state_object:
                obj.full_state_object.set_fullstate(True)
                obj.full_state_object.empty_pq(self.msg_type)

                #Indicate Full state to Netlink Monitor Thread
                obj.get_nlmonitor().set_fullstate(True)
            func(obj)
            if obj.full_state_object:
                #Indicate Full state stop to Netlink Monitor Thread
                obj.get_nlmonitor().set_fullstate(False)
                while obj.get_nlmonitor().get_fullstate_size():
                    time.sleep(self.fullstate_proc_delay)
                obj.full_state_object.set_fullstate(False)
                obj.full_state_object.flush_fullstate(self.msg_type)

        return wrapper

class FullStateMessages(enum.Enum):
    FULL_STATE = 0
    PARTIAL_STATE = 1
    FULL_STATE_SEGEMENT = 2
    FULL_STATE_COMPLETE = 3


class NetLinkFullStateObject(object):

    def __init__(self):
        self.__netlink_event = defaultdict(dict)
        self.__netlink_full_state = False
        self.pq = NetQPriorityQueue.PQ
        self.counter = {}

    def get_valid_msgtype(self, message_type):
        if message_type == "Routes":
            return "Route"
        else:
            return message_type

    def increment_counter(self, message_type):
        message_type = self.get_valid_msgtype(message_type)
        try:
            self.counter[message_type] += 1
        except Exception:
            self.counter[message_type] = 1

    def reset_counter(self, message_type):
        message_type = self.get_valid_msgtype(message_type)
        self.counter[message_type] = 0

    def get_reset_counter(self, message_type):
        message_type = self.get_valid_msgtype(message_type)
        try:
            counter = self.counter[message_type]
            self.reset_counter(message_type)
            return counter
        except Exception:
            return 0

    def set_fullstate(self, state):
        self.__netlink_full_state = state

    def get_fullstate(self):
        return self.__netlink_full_state

    def cache_netlink_buffer(self, obj, ifindex):
        self.__netlink_event[obj.message_type][ifindex] = obj

    def empty_pq(self, message_type):
        pq = NetQPriorityQueue.PQ

        priority = pq.get_pqid_by_message(message_type)
    
        pq.empty_pq(priority)


    def flush_fullstate(self, message_type):
        message = {}
        pq = NetQPriorityQueue.PQ

        message['message_type'] = message_type
        message['priority_qid'] = pq.get_pqid_by_message(message_type)
        message['counter'] = self.get_reset_counter(message_type)
        priority = self.__get_message_type_priority(message_type)

        try:
            queue_elem = QueueElem(priority, "FULL_STATE_END", True, message)
            pq.putq(queue_elem)
        except Exception:
            return False

        #* MACFDB generates VxLanRemoteDes */
        if(message_type == "MacFdb"):
            self.flush_fullstate("VxlanRemoteDest")


    def __get_message_type_priority(self, message_type):
        return 1

    def __reset_event_buff(self):
        self.__netlink_event = defaultdict(dict)

    def flush_netlink_buffer(self):

        pq = NetQPriorityQueue.PQ
        if pq is None:
            return False

        for message_type, interfaces in self.__netlink_event.iteritems():

            priority = self.__get_message_type_priority(message_type)

            message_data = [message.to_object_json() for ifidx, message in interfaces.iteritems()]

            try:
                queue_elem = QueueElem(priority, message_type, True, message_data)
                pq.putq(queue_elem)
            except Exception:
                return False

        self.__reset_event_buff()



NetlinkFullStateObject = NetLinkFullStateObject()
