# Copyright 2018, 2019 Cumulus Networks, Inc.  All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

"""ROM models
"""
import inspect
import time
import hashlib

from datetime import datetime
from netq_lib.common.configuration import is_cli_gateway
from netq_lib.orm import objects, HOSTNAME
from netq_lib.orm.memdb.models import MemModel
from netq_lib.common.nlfullstate import NetlinkFullStateObject
from netq_lib.orm.redisdb import Connection
from netq_lib.orm.redisdb.query import RedisQuery
from netq_lib.common.netq_pq import NetQPriorityQueue, QueueElem
from netq_lib.common.enums import ConfKw, ServiceStatus
from netq_lib.common.configuration import (
    get_config,
)

# If you add any classes to this file that is not an object pushed to the
# backend DB, such as  _RedisModel and Connection, please add those classes
# to the ignored list in agent.py's reinitialize_db() function.

_KEY_TTL = 90                   # TTL liveness is 90s


class _RedisModel(objects.NetQModel):
    __memmodel = None
    post_headers = dict()
    get_headers = dict()
    put_headers = dict()
    delete_headers = dict()
    agent_conf = get_config().get(ConfKw.NETQ_AGENT, dict())
    redis_enabled = agent_conf.get(ConfKw.IS_REDIS_ENABLED, ConfKw.DEFAULT_IS_REDIS_ENABLED)
    opta_enabled = agent_conf.get(ConfKw.IS_OPTA_ENABLED, ConfKw.DEFAULT_IS_OPTA_ENABLED)

    try:
        # TODO: completely decouple redis from api-gateway query layer once data validation is complete
        from netq_apps.lib.query import GatewayQuery
        query_cls = GatewayQuery if is_cli_gateway() else RedisQuery
    except ImportError:
        query_cls = RedisQuery

    _NETLINK_MSGS = ['Address', 'Route', 'Link', 'Netconf', 'Neighbor', 'MacFdb', 'VxlanRemoteDest']
    time_map = {'start': 'gt_timestamp', 'end': 'lt_timestamp', 'timestamp': 'eq_timestamp',
                'endtimestamp': 'lt_timestamp'}
    optional_args_map = dict()
    api_key_map = dict()
    api_filter_args = []
    app_name = ''
    delete_request_args = []
    post_request_args = []
    put_request_args = []

    active = objects.Boolean(json_key='active')
    hostname = objects.String(json_key='hostname')
    timestamp = objects.Float(json_key='timestamp')
    deleted = objects.Boolean(json_key='deleted')
    message_type = objects.String(json_key='messageType')

    def __init__(self, gateway=False):
        self.active = True
        self.deleted = False
        self.hostname = HOSTNAME
        self.timestamp = time.time()
        super(_RedisModel, self).__init__(gateway)
        # Fill the right message type after initialization
        self.message_type = self.__class__.__name__

    @classmethod
    def exclude_fmt(cls):
        """Return the list of fields to ignore when doing a comparison to
        find the differences between two instances of an object.
        See BgpSession for an example usage"""
        return cls.timestamp, cls.active

    @classmethod
    def memmodel(cls):
        if cls.__memmodel is None:
            cls.__memmodel = type(cls.__name__, (MemModel, cls), {})
        return cls.__memmodel

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return (cls.optional_args_map.get(key, key), val)
        else:
            return []

    @classmethod
    def key_fmt(cls, key_type=None):
        raise NotImplementedError

    @classmethod
    def score_fmt(cls, key_type=None):
        raise NotImplementedError

    @classmethod
    def val_fmt(cls):
        raise NotImplementedError

    @classmethod
    def url_fmt(cls, key=None):
        raise NotImplementedError

    def _get_connection_object(self, commit):
        if not self.redis_enabled:
            return
        if commit:
            return Connection.REDIS
        else:
            return Connection.PIPE

    @staticmethod
    def _get_queue_object():
        return NetQPriorityQueue.PQ

    def get_priority(self):
        return 1

    def copy(self, src):
        for _, field in inspect.getmembers(self.__class__):
            if (isinstance(field, objects.Field) and
                    field.name != self.__class__.timestamp.name):
                src_attribute = getattr(src, field.name, None)
                if src_attribute is not None:
                    setattr(self, field.name, src_attribute)

    def delete(self, purge=False):
        '''Mark the object as deleted OR purge the entry from the backend'''
        if purge:
            cls = self.__class__
            connection = self._get_connection_object(True)
            connection.zremrangebyrank(cls.obj_to_key(self.__dict__),
                                       min=0, max=-1)
            connection.hdel(cls.SEPERATOR.join((cls._namespace, cls.__name__)),
                            cls.obj_to_key(self.__dict__))
            # Keeping the local cached around in the case of the agent is
            # probably fine. See comment in flush_entry() below too.
        else:
            obj = self.__class__()
            obj.copy(self)
            obj.active = False
            obj.deleted = True
            obj.save()

    def flush_entry(self):
        '''purge a single entry from the sorted set with specified score'''
        cls = self.__class__
        score = cls.obj_to_score(self.__dict__)
        key = cls.obj_to_key(self.__dict__)
        return Connection.REDIS.zremrangebyscore(key, min=score, max=score)
        # From an agent perspective, the local cached entry is still around.
        # Is this the right thing or the wrong thing ? If I just deleted an
        # entry, do I want to add back the same entry again ? I'd think not.
        # By keeping the cached entry around, nothing gets pushed unless
        # something changes, which seems logical.

    def get_objid(self):
        '''A unique for this object type/hostname ID'''
        return hashlib.md5(self.obj_to_key(self.__dict__)).hexdigest()

    def __is_netlink_full_state_ongoing(self):

        if self.message_type not in self._NETLINK_MSGS:
            return False

        if (NetlinkFullStateObject.get_fullstate() is not True):
            return False
        else:
            return True

    def save_grpc(self):

        pq = self._get_queue_object()
        if pq is None:
            return False

        priority = self.get_priority()
        pq_data = self.to_object_json()
        if pq_data is None:
            return False

        if self.__is_netlink_full_state_ongoing():
            NetlinkFullStateObject.increment_counter(self.message_type)
            priority = pq.get_pqid_by_message(self.message_type)

        try:
            q_elem = QueueElem(priority, self.message_type, False, pq_data)
            pq.putq(q_elem)
        except Exception:
            # Raise Exception
            return False
        return True

    def save_entry(self, connection, publish=False):
        cls = self.__class__
        mcls = cls.memmodel()
        key = cls.obj_to_key(self.__dict__)
        val = cls.obj_to_val(self.__dict__)
        score = cls.obj_to_score(self.__dict__)

        add = True
        delete = False
        old_obj = None

        if mcls is not None:
            old_objs = (
                list(mcls.query.range(start=0, end=10000, first=False,
                                      keys=[key]))
            )
            if old_objs:
                old_obj = old_objs[-1]

                if self.equals(old_obj):
                    if old_obj.active == self.active:
                        add = False
                    elif self.active and len(old_objs) >= 2:
                        delete = True

                for obj in old_objs[:-2]:
                    obj.delete_object()

        if delete:
            del_score = cls.obj_to_score(old_obj.__dict__)
            mcls.delete(key, del_score)
            if self.redis_enabled:
                connection.zremrangebyscore(key, del_score, del_score)
        if add:
            mcls.save(key, score, val)
            if self.redis_enabled:
                connection.zadd(key, score, val)
                connection.hset(cls.SEPERATOR.join((
                    cls._namespace, cls.__name__)),
                    key, self.active)
                if publish:
                    connection.publish(key, val)
            if self.opta_enabled:
                self.save_grpc()

    def counter_fmt(self):
        return None

    def counter_retention_policy(self):
        return None

    def counter_continuous_query(self):
        return None

    def diff(self, other):
        '''
        Returns dict { k:(self,other) } differences between self and other
        instance same class. Only finds deltas between common keys
        '''

        if other is None:
            raise ValueError

        if not type(self) is type(other):
            raise NotImplementedError

        # WARNING: Using PREFIX from objects.
        exclude_fields = {'i_' + field.name for field in self.exclude_fmt()}
        return {k: (v, other.__dict__.get(k))
                for k, v in self.__dict__.items()
                if k not in exclude_fields and other.__dict__.get(k) != v}


class Netconf(_RedisModel):
    ifname = objects.String(json_key='interface')
    ifindex = objects.Integer(json_key='ifIndex')
    is_ipv6 = objects.Boolean(json_key='isIPv6')
    forwarding = objects.Boolean(json_key='forwarding')
    app_name = 'netconf'

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'ifname']

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return (cls.optional_args_map.get(key, key), val)
        else:
            return ['is_ipv6']

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.ifname, cls.is_ipv6)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return cls.ifindex, cls.forwarding, cls.timestamp, cls.active, cls.deleted

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(Netconf, self).save_entry(connection)


class Address(_RedisModel):
    ifname = objects.String(json_key='interface')
    prefix = objects.String(json_key='prefix')
    mask = objects.Integer(json_key='mask')
    is_ipv6 = objects.Boolean(json_key='isIPv6')
    vrf = objects.String(json_key='vrf')
    app_name = 'address'
    api_key_map = {'prefix': 'prefix'}

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return (cls.optional_args_map.get(key, key), val)
        else:
            return ['is_ipv6']

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'ifname', 'prefix', 'vrf']

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.ifname, cls.prefix,
                cls.mask, cls.is_ipv6, cls.vrf)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return cls.message_type, cls.timestamp, cls.active, cls.deleted

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(Address, self).save_entry(connection)


# TODO: class Inventory (notice the missing _) will replace all of these.
class _Inventory(_RedisModel):
    app_name = 'inventory'

    @classmethod
    def key_fmt(cls, key_type=None):
        raise NotImplementedError

    @classmethod
    def score_fmt(cls, key_type=None):
        raise NotImplementedError

    @classmethod
    def val_fmt(cls):
        raise NotImplementedError

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'ifname', 'type']


class ASIC(_Inventory):
    vendor = objects.String(json_key='vendor')
    model = objects.String(json_key='model')
    model_id = objects.String(json_key='modelId')
    core_bw = objects.String(json_key='coreBW')
    ports = objects.String(json_key='ports')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.vendor, cls.model,
                cls.model_id, cls.core_bw, cls.ports)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.timestamp,)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(ASIC, self).save_entry(connection, publish=True)


class BgpSession(_RedisModel):
    """Store BGP Session specific info
    """
    peer_name = objects.String(json_key='peerName')
    state = objects.String(json_key='peerState')
    peer_router_id = objects.String(json_key='bgpRouterId')
    peer_asn = objects.Long(json_key='peerAsn')
    peer_hostname = objects.String(json_key='peerHostname')
    asn = objects.Long(json_key='bgpAsn')
    reason = objects.String(json_key='reason')
    ipv4_pfx_rcvd = objects.Integer(json_key='ipv4PrefixRx')
    ipv6_pfx_rcvd = objects.Integer(json_key='ipv6PrefixRx')
    evpn_pfx_rcvd = objects.Integer(json_key='evpnPrefixRx')
    last_reset_time = objects.Float(json_key='lastResetTime')
    up_time = objects.Float(json_key='upTime')
    conn_estd = objects.Integer(json_key='connEstd')
    conn_dropped = objects.Integer(json_key='connDropped')
    upd8_rx = objects.Integer(json_key='upd8Rx')
    upd8_tx = objects.Integer(json_key='upd8Tx')
    vrfid = objects.Integer(json_key='vrfId')
    vrf = objects.String(json_key='vrf')
    tx_families = objects.Json(json_key='txFamilies')
    rx_families = objects.Json(json_key='rxFamilies')
    objid = objects.String(json_key='objId')
    api_key_map = {'peer_name': 'peername'}
    app_name = 'bgp'

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'peer_name', 'asn', 'vrf']

    @classmethod
    def exclude_fmt(cls):
        return (cls.timestamp, cls.active, cls.last_reset_time, cls.up_time,
                cls.upd8_tx, cls.upd8_rx, cls.evpn_pfx_rcvd, cls.ipv4_pfx_rcvd,
                cls.ipv6_pfx_rcvd)

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.peer_name, cls.asn, cls.vrf

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.message_type, cls.state, cls.peer_router_id, cls.peer_asn, cls.peer_hostname,
                cls.reason, cls.ipv4_pfx_rcvd, cls.ipv6_pfx_rcvd,
                cls.evpn_pfx_rcvd, cls.timestamp, cls.last_reset_time,
                cls.conn_estd, cls.conn_dropped, cls.upd8_rx, cls.vrfid,
                cls.upd8_tx, cls.up_time, cls.tx_families, cls.objid,
                cls.rx_families, cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(BgpSession, self).save_entry(connection, publish=True)


class Board(_Inventory):
    vendor = objects.String(json_key='vendor')
    model = objects.String(json_key='model')
    base_mac = objects.String(json_key='baseMac')
    part_number = objects.String(json_key='partNo')
    mfg_date = objects.String(json_key='mfgDate')
    serial_number = objects.String(json_key='serialNo')
    label_revision = objects.String(json_key='rev')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.vendor, cls.model,
                cls.base_mac, cls.part_number, cls.mfg_date, cls.serial_number,
                cls.label_revision)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.message_type, cls.timestamp,)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(Board, self).save_entry(connection, publish=True)


class ClagSession(_RedisModel):
    role = objects.String(json_key='role')
    peer_role = objects.String(json_key='peerRole')
    peer_state = objects.Boolean(json_key='peerState')
    peer_if = objects.String(json_key='clagPeerIf')
    clag_sysmac = objects.String(json_key='sysMac')
    backup_ip_active = objects.Boolean(json_key='backupIpActive')
    backup_ip = objects.String(json_key='backupIp')
    single_bonds = objects.Json(json_key='singleBonds')
    dual_bonds = objects.Json(json_key='dualAttachedBonds')
    conflicted_bonds = objects.Json(json_key='conflictedBonds')
    proto_down_bonds = objects.Json(json_key='protoDownBonds')
    vxlan_anycast = objects.String(json_key='vxlanAnycast')
    api_key_map = {'clag_sysmac': 'sysmac'}
    app_name = 'clag'

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'clag_sysmac']

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.clag_sysmac

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.message_type, cls.peer_role, cls.role, cls.peer_state, cls.peer_if,
                cls.backup_ip, cls.backup_ip_active, cls.single_bonds,
                cls.dual_bonds, cls.timestamp, cls.conflicted_bonds,
                cls.vxlan_anycast, cls.proto_down_bonds, cls.active,
                cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(ClagSession, self).save_entry(connection, publish=True)


class Command(_RedisModel):
    key = objects.String(json_key='key')
    output = objects.String(json_key='output')

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.key

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return cls.output,

    def save(self, commit=False):
        if self.redis_enabled:
            connection = self._get_connection_object(commit)
            connection.zadd(Command.obj_to_key(self.__dict__),
                            Command.obj_to_score(self.__dict__),
                            Command.obj_to_val(self.__dict__))
            connection.hset(self.__class__.SEPERATOR.join((
                self.__class__._namespace,
                self.__class__.__name__)),
                Command.obj_to_key(self.__dict__),
                True)
        if self.opta_enabled:
            self.save_grpc()


class CPU(_Inventory):
    arch = objects.String(json_key='arch')
    nos = objects.String(json_key='cores')
    model = objects.String(json_key='model')
    max_freq = objects.String(json_key='freq')
    mem_total = objects.String(json_key='memory')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.arch, cls.nos, cls.model,
                cls.max_freq, cls.mem_total)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.timestamp,)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(CPU, self).save_entry(connection, publish=True)


class Description(_RedisModel):
    objtype = objects.String(json_key='objType')
    descrid = objects.String(json_key='descrId')
    description = objects.String(json_key='description')

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.objtype, cls.descrid

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.description, cls.timestamp, cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(Description, self).save_entry(connection, publish=False)


class Disk(_Inventory):
    name = objects.String(json_key='name')
    size = objects.String(json_key='size')
    d_type = objects.String(json_key='type')
    vendor = objects.String(json_key='vendor')
    transport = objects.String(json_key='transport')
    rev = objects.String(json_key='rev')
    model = objects.String(json_key='model')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.name, cls.size, cls.d_type,
                cls.vendor, cls.transport, cls.rev, cls.model)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.timestamp,)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(Disk, self).save_entry(connection, publish=True)


class DockerContainer(_RedisModel):
    container_id = objects.String(json_key='containerId')
    gw = objects.String(json_key='gateway')
    image = objects.String(json_key='image')
    ip = objects.String(json_key='containerIp')
    mac = objects.String(json_key='mac')
    name = objects.String(json_key='containerName')
    network_name = objects.String(json_key='networkName')
    network_id = objects.String(json_key='networkId')
    prefix_len = objects.Integer(json_key='prefixLen')
    port_list = objects.String(json_key='portList')
    service_id = objects.String(json_key='serviceId')
    service_name = objects.String(json_key='serviceName')
    status = objects.String(json_key='status')
    start_time = objects.String(json_key='startTime')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname,
                cls.name, cls.image,
                cls.network_name, cls.ip, cls.mac,
                cls.service_name)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.timestamp, cls.container_id, cls.status,
                cls.network_id, cls.gw, cls.prefix_len,
                cls.port_list, cls.service_id,
                cls.start_time, cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(DockerContainer, self).save_entry(connection)


class DockerPortMap(_RedisModel):
    container_port = objects.String(json_key='containerPort')
    container_id = objects.String(json_key='containerId')
    container_ip = objects.String(json_key='containerIp')
    host_port = objects.String(json_key='hostPort')
    host_ip = objects.String(json_key='hostIp')
    image = objects.String(json_key='image')
    mac = objects.String(json_key='mac')
    name = objects.String(json_key='containerName')
    network_name = objects.String(json_key='networkName')
    network_id = objects.String(json_key='networkId')
    node_id = objects.String(json_key='nodeId')
    proto = objects.String(json_key='proto')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname,
                cls.name, cls.container_ip, cls.proto, cls.container_port,
                cls.host_ip, cls.host_port, cls.network_name)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.timestamp, cls.container_id, cls.network_id,
                cls.image, cls.mac, cls.node_id, cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(DockerPortMap, self).save_entry(connection)


class DockerSwarmCluster(_RedisModel):
    cluster_version = objects.String(json_key='clusterVersion')
    cluster_name = objects.String(json_key='clusterName')
    cluster_id = objects.String(json_key='clusterId')
    docker_version = objects.String(json_key='version')
    nodes = objects.String(json_key='workerNodes')
    num_nodes = objects.Integer(json_key='numNodes')
    num_managers = objects.Integer(json_key='numManagers')
    managers = objects.String(json_key='managerNodes')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.cluster_name)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.docker_version, cls.cluster_version, cls.cluster_id,
                cls.num_nodes, cls.num_managers, cls.managers,
                cls.timestamp, cls.nodes, cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(DockerSwarmCluster, self).save_entry(connection)


class DockerSwarmNode(_RedisModel):
    cluster_id = objects.String(json_key='clusterId')
    cluster_name = objects.String(json_key='clusterName')
    docker_version = objects.String(json_key='dockerVersion')
    node_id = objects.String(json_key='nodeId')
    node_name = objects.String(json_key='nodeName')
    node_state = objects.String(json_key='nodeState')
    plugins = objects.String(json_key='plugins')
    role = objects.String(json_key='role')
    availability = objects.String(json_key='availability')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.cluster_name, cls.node_name)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.timestamp, cls.docker_version, cls.cluster_id, cls.node_id,
                cls.node_state, cls.role, cls.plugins,
                cls.availability, cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(DockerSwarmNode, self).save_entry(connection)


class DockerNetwork(_RedisModel):
    container_list = objects.String(json_key='containerList')
    driver = objects.String(json_key='driver')
    host_binding = objects.String(json_key='hostBinding')
    ipv6_enabled = objects.Boolean(json_key='ipv6Enabled')
    ip_masquerade = objects.Boolean(json_key='ipMasq')
    ipam_driver = objects.String(json_key='ipamDriver')
    encrypted = objects.Boolean(json_key='encrypted')
    default_bridge = objects.Boolean(json_key='defaultBridge')
    network_name = objects.String(json_key='networkName')
    network_id = objects.String(json_key='networkId')
    subnet = objects.String(json_key='subnet')
    gateway = objects.String(json_key='gateway')
    parent_interface = objects.String(json_key='parentInterface')
    vxlan_id = objects.Integer(json_key='vxlanId')
    mtu = objects.Integer(json_key='mtu')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.network_name, cls.driver)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.gateway, cls.parent_interface, cls.vxlan_id, cls.network_id,
                cls.mtu, cls.host_binding, cls.ipv6_enabled, cls.ip_masquerade,
                cls.encrypted, cls.default_bridge, cls.ipam_driver, cls.subnet,
                cls.container_list, cls.timestamp, cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(DockerNetwork, self).save_entry(connection)


class DockerService(_RedisModel):
    image = objects.String(json_key='image')
    mode = objects.String(json_key='mode')
    parallelism = objects.Integer(json_key='parallelism')
    port_list = objects.String(json_key='portMapping')
    network_list = objects.String(json_key='networkList')
    replicas = objects.Integer(json_key='replicas')
    service_name = objects.String(json_key='serviceName')
    service_id = objects.String(json_key='serviceId')
    version = objects.Integer(json_key='version')
    vip = objects.String(json_key='virtualIp')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname,
                cls.service_name, cls.mode)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.image, cls.replicas, cls.parallelism, cls.service_id,
                cls.port_list, cls.network_list, cls.vip, cls.version,
                cls.timestamp, cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(DockerService, self).save_entry(connection)


class DockerHost(_RedisModel):
    containers = objects.Integer(json_key='installed')
    containers_running = objects.Integer(json_key='running')
    docker_version = objects.String(json_key='version')
    images = objects.Integer(json_key='images')
    ip_forwarding = objects.Boolean(json_key='ipForwarding')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.docker_version)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.images, cls.containers, cls.containers_running,
                cls.ip_forwarding, cls.timestamp, cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(DockerHost, self).save_entry(connection)


class EsxHost(_RedisModel):
    vcenter = objects.String(json_key='vcenter')
    datacenter = objects.String(json_key='datacenter')
    cluster = objects.String(json_key='cluster')
    esxhost = objects.String(json_key='esxhost')
    pnic = objects.String(json_key='pnic')
    dvs = objects.String(json_key='dvs')
    pg = objects.String(json_key='pg')
    encaptype = objects.Integer(json_key='encaptype')
    encap = objects.Integer(json_key='encap')
    vmname = objects.String(json_key='vmname')
    ip = objects.String(json_key='ip')
    mac = objects.String(json_key='mac')
    mtu = objects.Integer(json_key='mtu')
    pnic_admin_state = objects.String(json_key='pnic_admin_state')
    pnic_oper_state = objects.String(json_key='pnic_oper_state')
    peer_hostname = objects.String(json_key='peer_hostname')
    peer_ifname = objects.String(json_key='peer_ifname')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname,
                cls.vcenter, cls.datacenter, cls.cluster, cls.esxhost,
                cls.pnic, cls.dvs, cls.pg, cls.encaptype, cls.encap)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.ip, cls.mac, cls.mtu, cls.pnic_admin_state,
                cls.pnic_oper_state, cls.peer_hostname, cls.peer_ifname,
                cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(EsxHost, self).save_entry(connection)


class EsxVmAttr(_RedisModel):
    vcenter = objects.String(json_key='vcenter')
    datacenter = objects.String(json_key='datacenter')
    vmname = objects.String(json_key='vmname')
    ostype = objects.String(json_key='ostype')
    vnic = objects.String(json_key='vnic')
    pg = objects.String(json_key='portgroup')
    ip = objects.String(json_key='ip')
    mac = objects.String(json_key='mac')
    esxhost = objects.String(json_key='esxhost')
    mem = objects.String(json_key='memory')
    disk = objects.String(json_key='disk')
    cpus = objects.String(json_key='numcpus')
    vmpath = objects.String(json_key='vmpath')
    state = objects.String(json_key='powerstate')
    annotation = objects.String(json_key='annotation')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.vcenter, cls.datacenter,
                cls.vmname, cls.ostype, cls.vnic, cls.pg, cls.ip, cls.mac)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.esxhost, cls.mem, cls.disk, cls.cpus, cls.vmpath,
                cls.state, cls.annotation, cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(EsxVmAttr, self).save_entry(connection)


class EsxVmBind(_RedisModel):
    vcenter = objects.String(json_key='vcenter')
    datacenter = objects.String(json_key='datacenter')
    cluster = objects.String(json_key='cluster')
    esxhost = objects.String(json_key='esxhost')
    vmname = objects.String(json_key='vmname')
    ostype = objects.String(json_key='ostype')
    state = objects.String(json_key='state')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.vcenter, cls.datacenter,
                cls.cluster, cls.esxhost, cls.vmname, cls.ostype, cls.state)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.state, cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(EsxVmBind, self).save_entry(connection)


class EvpnVniInfo(_RedisModel):
    '''Contains VNI info obtained from bgp evpn'''
    vni = objects.Integer(json_key='vni')
    origin_ip = objects.String(json_key='originIp')
    rd = objects.String(json_key='rd')
    export_rt = objects.Json(json_key='exportRT')
    import_rt = objects.Json(json_key='importRT')
    in_kernel = objects.Boolean(json_key='inKernel')
    adv_all_vni = objects.Boolean(json_key='advAllVni')
    adv_gw_ip = objects.String(json_key='advGatewayMacIp')
    is_l3 = objects.Boolean(json_key='isL3')
    app_name = 'evpn'

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.vni

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.origin_ip, cls.rd, cls.export_rt, cls.import_rt,
                cls.in_kernel, cls.timestamp, cls.adv_all_vni, cls.adv_gw_ip,
                cls.active, cls.deleted, cls.is_l3)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(EvpnVniInfo, self).save_entry(connection, publish=True)

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'vni']


class _Sensor(_RedisModel):
    app_name = 'sensor'
    api_key_map = {'s_name': 'name'}

    @classmethod
    def key_fmt(cls, key_type=None):
        raise NotImplementedError

    @classmethod
    def score_fmt(cls, key_type=None):
        raise NotImplementedError

    @classmethod
    def val_fmt(cls):
        raise NotImplementedError

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'type', 's_name']


class Fan(_Sensor):
    message_type = objects.String(json_key='messageType')
    s_name = objects.String(json_key='name')
    s_desc = objects.String(json_key='description')
    s_input = objects.Integer(json_key='speed')
    s_max = objects.Integer(json_key='max')
    s_min = objects.Integer(json_key='min')
    s_state = objects.String(json_key='state')
    s_prev_state = objects.String(json_key='prevState')
    s_msg = objects.String(json_key='message')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.s_name,
                cls.s_desc)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.timestamp, cls.s_state, cls.s_prev_state, cls.s_input,
                cls.s_msg, cls.s_max, cls.s_min, cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(Fan, self).save_entry(connection, publish=True)
        # pq = self._get_queue_object()
        # priority = self.get_priority()
        # pq_data = self.to_json()
        # q_elem = QueueElem(priority, pq_data)
        # pq.putq(q_elem)


class Inventory(_RedisModel):
    agent_version = objects.String(json_key='agent_version')
    asic_vendor = objects.String(json_key='asic_vendor')
    asic_model = objects.String(json_key='asic_model')
    asic_model_id = objects.String(json_key='asic_modelId')
    asic_core_bw = objects.String(json_key='asic_coreBW')
    asic_ports = objects.String(json_key='asic_ports')
    asic_data = objects.String(json_key='asic_data')
    platform_vendor = objects.String(json_key='platform_vendor')
    platform_model = objects.String(json_key='platform_model')
    platform_base_mac = objects.String(json_key='platform_baseMac')
    platform_part_number = objects.String(json_key='platform_partNo')
    platform_mfg_date = objects.String(json_key='platform_mfgDate')
    platform_serial_number = objects.String(json_key='platform_serialNo')
    platform_label_revision = objects.String(json_key='platform_rev')
    cpu_arch = objects.String(json_key='cpu_arch')
    cpu_nos = objects.String(json_key='cpu_cores')
    cpu_model = objects.String(json_key='cpu_model')
    cpu_max_freq = objects.String(json_key='cpu_freq')
    cpu_data = objects.String(json_key='cpu_data')
    disk_total_size = objects.String(json_key='disk_size')
    disk_data = objects.String(json_key='disk_data')
    license_state = objects.String(json_key='license_state')
    license_data = objects.String(json_key='license_data')
    memory_total_size = objects.String(json_key='memory_size')
    memory_data = objects.String(json_key='memory_data')
    os_name = objects.String(json_key='os_name')
    os_version = objects.String(json_key='os_fullVersion')
    os_version_id = objects.String(json_key='os_version')
    app_name = 'inventory'
    api_filter_args = ['hostname', 'asic_vendor', 'asic_model', 'asic_model_id', 'platform_vendor', 'platform_model',
                       'cpu_arch', 'disk_name', 'disk_transport', 'disk_vendor', 'license_name', 'license_state',
                       'memory_vendor', 'memory_m_type', 'os_name', 'os_version']

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname

    @classmethod
    def val_fmt(cls, key_type=None):
        return (cls.asic_vendor, cls.asic_model, cls.asic_model_id, cls.asic_core_bw, cls.asic_ports, cls.asic_data,
                cls.platform_vendor, cls.platform_model, cls.platform_base_mac, cls.platform_part_number,
                cls.platform_mfg_date, cls.platform_serial_number, cls.platform_label_revision,
                cls.cpu_arch, cls.cpu_nos, cls.cpu_model, cls.cpu_max_freq, cls.cpu_data,
                cls.disk_total_size, cls.disk_data,
                cls.license_state, cls.license_data,
                cls.memory_total_size, cls.memory_data,
                cls.os_name, cls.os_version, cls.os_version_id, cls.agent_version,
                cls.message_type, cls.timestamp)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(Inventory, self).save_entry(connection, publish=True)

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['type']


class KubeCluster(_RedisModel):
    api_server = objects.String(json_key='api_server')
    cluster_name = objects.String(json_key='cluster_name')
    nodes = objects.String(json_key='nodes')
    controller_manager = objects.String(json_key='controller_manager')
    scheduler = objects.String(json_key='scheduler')
    etcd = objects.String(json_key='etcd')
    app_name = 'kubernetes'
    api_key_map = {'cluster_name': 'clustername'}

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['path', 'hostname', 'api_server', 'cluster_name']

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.api_server, cls.cluster_name

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.nodes, cls.timestamp,
                cls.controller_manager,
                cls.scheduler, cls.etcd,
                cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(KubeCluster, self).save_entry(connection, publish=True)


class KubeDeployment(_RedisModel):
    api_server = objects.String(json_key='api_server')
    cluster_name = objects.String(json_key='cluster_name')
    namespace = objects.String(json_key='namespace')
    name = objects.String(json_key='name')
    labels = objects.String(json_key='labels')
    replicas = objects.Integer(json_key='replicas')
    ready_replicas = objects.Integer(json_key='ready_replicas')
    start_time = objects.String(json_key='start_time')
    app_name = 'kubernetes'
    api_key_map = {'cluster_name': 'clustername'}

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['path', 'hostname', 'cluster_name', 'namespace', 'name']

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return (cls.optional_args_map.get(key, key), val)
        else:
            return ['label']

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.api_server,
                cls.cluster_name, cls.namespace, cls.name)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.timestamp, cls.labels,
                cls.replicas, cls.ready_replicas, cls.start_time,
                cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(KubeDeployment, self).save_entry(connection, publish=True)


class KubeDaemonSet(_RedisModel):
    api_server = objects.String(json_key='api_server')
    cluster_name = objects.String(json_key='cluster_name')
    namespace = objects.String(json_key='namespace')
    name = objects.String(json_key='name')
    labels = objects.String(json_key='labels')
    desired_count = objects.Integer(json_key='desired_count')
    ready_count = objects.Integer(json_key='ready_count')
    start_time = objects.String(json_key='start_time')
    app_name = 'kubernetes'
    api_key_map = {'cluster_name': 'clustername'}

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['path', 'hostname', 'cluster_name', 'namespace', 'name']

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return (cls.optional_args_map.get(key, key), val)
        else:
            return ['label']

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.api_server,
                cls.cluster_name, cls.namespace, cls.name)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.timestamp, cls.labels,
                cls.desired_count, cls.ready_count, cls.start_time,
                cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(KubeDaemonSet, self).save_entry(connection, publish=True)


class KubeNode(_RedisModel):
    api_server = objects.String(json_key='api_server')
    cluster_name = objects.String(json_key='cluster_name')
    node_name = objects.String(json_key='node_name')
    labels = objects.String(json_key='labels')
    role = objects.String(json_key='role')
    pod_cidr = objects.String(json_key='pod_cidr')
    kube_proxy_version = objects.String(json_key='kube_proxy_version')
    kubelet_version = objects.String(json_key='kubelet_version')
    status = objects.String(json_key='status')
    machine_id = objects.String(json_key='machine_id')
    container_runtime = objects.String(json_key='container_runtime')
    start_time = objects.String(json_key='start_time')
    app_name = 'kubernetes'
    api_key_map = {'cluster_name': 'clustername'}

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['path', 'hostname', 'cluster_name', 'namespace', 'name']

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return (cls.optional_args_map.get(key, key), val)
        else:
            return ['label']

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.api_server,
                cls.cluster_name, cls.node_name)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.role, cls.labels, cls.pod_cidr,
                cls.kube_proxy_version, cls.kubelet_version,
                cls.machine_id, cls.status, cls.container_runtime,
                cls.timestamp, cls.start_time,
                cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(KubeNode, self).save_entry(connection, publish=True)


class KubePod(_RedisModel):
    api_server = objects.String(json_key='api_server')
    cluster_name = objects.String(json_key='cluster_name')
    namespace = objects.String(json_key='namespace')
    name = objects.String(json_key='name')
    labels = objects.String(json_key='labels')
    template_hash = objects.String(json_key='template_hash')
    node_name = objects.String(json_key='node_name')
    containers = objects.String(json_key='containers')
    ip_address = objects.String(json_key='ip_address')
    owner_kind = objects.String(json_key='owner_kind')
    owner_name = objects.String(json_key='owner_name')
    start_time = objects.String(json_key='start_time')
    status = objects.String(json_key='status')
    app_name = 'kubernetes'
    api_key_map = {'cluster_name': 'clustername'}
    optional_args_map = {'ip_address': 'podip', 'node_name': 'nodename'}

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['path', 'hostname', 'cluster_name', 'namespace', 'name']

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return (cls.optional_args_map.get(key, key), val)
        else:
            return ['label', 'ip_address', 'node_name']

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.api_server,
                cls.cluster_name,
                cls.namespace, cls.name)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.timestamp, cls.labels, cls.template_hash,
                cls.node_name, cls.containers, cls.ip_address,
                cls.owner_kind, cls.owner_name, cls.start_time,
                cls.status, cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(KubePod, self).save_entry(connection, publish=True)


class KubeReplicationController(_RedisModel):
    api_server = objects.String(json_key='api_server')
    cluster_name = objects.String(json_key='cluster_name')
    namespace = objects.String(json_key='namespace')
    name = objects.String(json_key='name')
    labels = objects.String(json_key='labels')
    replicas = objects.Integer(json_key='replicas')
    ready_replicas = objects.Integer(json_key='ready_replicas')
    start_time = objects.String(json_key='start_time')
    app_name = 'kubernetes'
    api_key_map = {'cluster_name': 'clustername'}

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['path', 'hostname', 'cluster_name', 'namespace', 'name']

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return (cls.optional_args_map.get(key, key), val)
        else:
            return ['label']

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.api_server,
                cls.cluster_name, cls.namespace, cls.name)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.timestamp, cls.labels,
                cls.replicas, cls.ready_replicas, cls.start_time,
                cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(KubeReplicationController, self).save_entry(connection, publish=True)


class KubeReplicaSet(_RedisModel):
    api_server = objects.String(json_key='api_server')
    cluster_name = objects.String(json_key='cluster_name')
    namespace = objects.String(json_key='namespace')
    name = objects.String(json_key='name')
    labels = objects.String(json_key='labels')
    pod_template_hash = objects.String(json_key='pod_template_hash')
    replicas = objects.Integer(json_key='replicas')
    ready_replicas = objects.Integer(json_key='ready_replicas')
    start_time = objects.String(json_key='start_time')
    app_name = 'kubernetes'
    api_key_map = {'cluster_name': 'clustername'}

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['path', 'hostname', 'cluster_name', 'namespace', 'name']

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return (cls.optional_args_map.get(key, key), val)
        else:
            return ['label']

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.api_server,
                cls.cluster_name, cls.namespace, cls.name)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.timestamp, cls.labels,
                cls.replicas, cls.ready_replicas, cls.pod_template_hash,
                cls.start_time, cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(KubeReplicaSet, self).save_entry(connection, publish=True)


class KubeService(_RedisModel):
    api_server = objects.String(json_key='api_server')
    cluster_name = objects.String(json_key='cluster_name')
    namespace = objects.String(json_key='namespace')
    name = objects.String(json_key='name')
    labels = objects.String(json_key='labels')
    service_type = objects.String(json_key='service_type')
    cluster_ip = objects.String(json_key='cluster_ip')
    external_ip = objects.String(json_key='external_ip')
    ports = objects.String(json_key='port')
    status = objects.String(json_key='status')
    start_time = objects.String(json_key='start_time')
    app_name = 'kubernetes'
    api_key_map = {'cluster_name': 'clustername'}
    optional_args_map = {'cluster_ip': 'clusterip', 'external_ip': 'externalip'}

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['path', 'hostname', 'cluster_name', 'namespace', 'name']

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return (cls.optional_args_map.get(key, key), val)
        else:
            return ['label', 'cluster_ip', 'external_ip']

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.api_server,
                cls.cluster_name, cls.namespace, cls.name)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.timestamp, cls.labels,
                cls.service_type, cls.cluster_ip,
                cls.external_ip, cls.ports, cls.status,
                cls.start_time, cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(KubeService, self).save_entry(connection, publish=True)


class License(_Inventory):
    name = objects.String(json_key='name')
    state = objects.String(json_key='state')
    license = objects.String(json_key='license')

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.name

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return cls.state, cls.license, cls.timestamp

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(License, self).save_entry(connection, publish=True)


class Link(_RedisModel):
    message_type = objects.String(json_key='messageType')
    ifname = objects.String(json_key='interface')
    ifindex = objects.Integer(json_key='ifIndex')
    mac_address = objects.MacAddress(json_key='macAddress')
    admin_state = objects.String(json_key='adminState')
    oper_state = objects.String(json_key='operState')
    managed = objects.Boolean(json_key='managed')
    kind = objects.String(json_key='type')
    mtu = objects.Integer(json_key='mtu')
    master = objects.String(json_key='master')
    is_vlan_filtering = objects.Boolean(json_key='isVlanFiltering')
    vlans = objects.String(json_key='vlans')
    access_vlan = objects.Integer(json_key='pvid')
    dstport = objects.Integer(json_key='dstPort')
    localip = objects.String(json_key='localIp')
    vni = objects.Integer(json_key='vni')
    vrf = objects.String(json_key='vrf')
    rt_table_id = objects.Integer(json_key='routeTableId')
    parent_if = objects.String(json_key='parentIf')
    stp_state = objects.Integer(json_key='stpState')
    down_reason = objects.String(json_key='downReason')
    learning_en = objects.Boolean(json_key='learningEnabled')
    arp_suppress_en = objects.Boolean(json_key='arpSuppressionEnabled')
    objid = objects.String(json_key='objId')
    app_name = 'link'

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'ifname', 'kind', 'master', 'vni', 'state']

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return (cls.optional_args_map.get(key, key), val)
        else:
            return []

    @classmethod
    def key_fmt(cls, key_type=None):
        """We need to query sometimes by VNI and so we stick VNI in the key.
        Its 0 for anything but Vxlan.
        """
        return (cls.__name__, cls.hostname, cls.ifname, cls.kind, cls.vni,
                cls.master)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.message_type, cls.admin_state, cls.oper_state, cls.managed, cls.mtu,
                cls.ifindex, cls.is_vlan_filtering, cls.timestamp,
                cls.vlans, cls.access_vlan, cls.localip, cls.down_reason,
                cls.vrf, cls.rt_table_id, cls.parent_if, cls.stp_state,
                cls.mac_address, cls.dstport, cls.learning_en, cls.objid,
                cls.arp_suppress_en, cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(Link, self).save_entry(connection, publish=True)


class Liveness(_RedisModel):

    def __str__(self):
        return self.hostname

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return cls.hostname,

    def save(self, commit=False):
        if self.redis_enabled:
            connection = self._get_connection_object(commit)
            key = Liveness.obj_to_key(self.__dict__)
            val = Liveness.obj_to_val(self.__dict__)
            score = Liveness.obj_to_score(self.__dict__)
            connection.zadd(key, score, val)
            connection.expire(key, _KEY_TTL)
            connection.hset(self.__class__.SEPERATOR.join((
                self.__class__._namespace,
                self.__class__.__name__)),
                key, True)
        if self.opta_enabled:
            self.save_grpc()


class ConfigMon(_RedisModel):
    file_name = objects.String(json_key='fileName')
    file_content = objects.String(json_key='fileContent')
    last_modify_time = objects.Float(json_key='lastModifyTime')
    app_name = 'configmon'

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'file_name']

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.file_name

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.message_type, cls.file_name, cls.file_content, cls.timestamp, cls.last_modify_time,
                cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(ConfigMon, self).save_entry(connection, publish=True)


class ClSupport(_RedisModel):
    file_abs_name = objects.String(json_key='fileAbsName')
    create_time = objects.Float(json_key='createTime')
    app_name = 'clsupport'

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'file_abs_name']

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.file_abs_name

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return cls.message_type, cls.file_abs_name, cls.create_time, cls.timestamp, cls.active, cls.deleted

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(ClSupport, self).save_entry(connection, publish=True)


class AgentStats(_RedisModel):
    agent_util = objects.Json(json_key='agentUtil')
    grpc_util = objects.Json(json_key='grpcUtil')
    message_count = objects.Json(json_key='messageCount')
    msg_type_count = objects.Json(json_key='msgTypeCount')
    msgtype_fullstate_count = objects.Json(json_key='msgtypeFullstateCcount')
    pq_full_count = objects.Integer(json_key='pqFullCount')
    pq_util = objects.Json(json_key='pqUtil')
    agent_pause_count = objects.Integer(json_key='agentPauseCount')
    chunk_count = objects.Json(json_key='chunkCount')
    opta_grps_reset_count = objects.Integer(json_key='optaGrpsResetCount')
    app_name = 'agentstats'

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname']

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.message_type, cls.message_count, cls.msg_type_count, cls.msgtype_fullstate_count, cls.grpc_util,
                cls.agent_util, cls.pq_full_count, cls.pq_util, cls.agent_pause_count, cls.chunk_count,
                cls.opta_grps_reset_count, cls.timestamp, cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(AgentStats, self).save_entry(connection, publish=True)


class ResourceUtil(_RedisModel):
    cpu_utilization = objects.Float(json_key='cpuUtilization')
    mem_utilization = objects.Float(json_key='memUtilization')
    app_name = 'resource'

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname']

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return cls.message_type, cls.cpu_utilization, cls.mem_utilization, cls.timestamp, cls.active, cls.deleted

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(ResourceUtil, self).save_entry(connection, publish=True)


class Lldp(_RedisModel):
    ifname = objects.String(json_key='interface')
    peer_hostname = objects.String(json_key='peerHostname')
    peer_ifname = objects.String(json_key='peerInterface')
    lldp_peer_bridge = objects.Boolean(json_key='lldpPeerBridge')
    lldp_peer_router = objects.Boolean(json_key='lldpPeerRouter')
    lldp_peer_station = objects.Boolean(json_key='lldpPeerStation')
    lldp_peer_os = objects.String(json_key='lldpPeerOS')
    lldp_peer_osv = objects.String(json_key='lldpPeerOsVersion')
    app_name = 'lldp'

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'ifname']

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.ifname, cls.peer_hostname

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.message_type, cls.peer_ifname, cls.lldp_peer_bridge, cls.lldp_peer_router,
                cls.lldp_peer_station, cls.lldp_peer_os, cls.lldp_peer_osv,
                cls.timestamp, cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(Lldp, self).save_entry(connection, publish=True)



class Ptm(_RedisModel):
    cbl_status = objects.String(json_key='cblStatus')
    port = objects.String(json_key='port')
    exp_nbr = objects.String(json_key='expNbr')
    act_nbr = objects.String(json_key='actNbr')

    app_name = 'ptm'

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'port']

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.port

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.message_type, cls.cbl_status, cls.port, cls.exp_nbr, cls.act_nbr,
                cls.timestamp, cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(Ptm, self).save_entry(connection, publish=True)


class LnvSession(_RedisModel):
    role = objects.String(json_key='role')
    snd_ip = objects.String(json_key='sndIp')
    repl_mode = objects.String(json_key='replicationMode')
    rd_peers = objects.String(json_key='vxrdPeers')
    snd_peers = objects.Json(json_key='vxsndPeers')
    vnis = objects.Json(json_key='vnis')
    state = objects.String(json_key='state')
    version = objects.String(json_key='version')
    src_ip = objects.String(json_key='srcIp')
    app_name = 'lnv'

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.role

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'role']

    @classmethod
    def val_fmt(cls):
        return (cls.role, cls.snd_ip, cls.rd_peers, cls.snd_peers,
                cls.vnis, cls.state, cls.repl_mode, cls.version,
                cls.active, cls.deleted, cls.timestamp, cls.src_ip)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(LnvSession, self).save_entry(connection, publish=True)


class MacFdb(_RedisModel):
    mac_address = objects.MacAddress(json_key='macAddress')
    origin = objects.Boolean(json_key='origin')
    nexthop = objects.String(json_key='egressPort')
    vlan = objects.Integer(json_key='vlan')
    dst = objects.String(json_key='dst')
    port = objects.String(json_key='port')
    is_remote = objects.Boolean(json_key='isRemote')
    is_static = objects.Boolean(json_key='isStatic')
    api_key_map = {'mac_address': 'macaddress'}
    app_name = 'macfdb'

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'mac_address', 'vlan']

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.mac_address, cls.vlan

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.origin, cls.nexthop, cls.dst, cls.port, cls.timestamp,
                cls.is_remote, cls.is_static, cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(MacFdb, self).save_entry(connection)


class Memory(_Inventory):
    vendor = objects.String(json_key='vendor')
    size = objects.String(json_key='size')
    speed = objects.String(json_key='speed')
    name = objects.String(json_key='name')
    m_type = objects.String(json_key='type')
    serial_number = objects.String(json_key='serialNo')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.name, cls.size, cls.speed,
                cls.m_type, cls.vendor, cls.serial_number)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.timestamp,)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(Memory, self).save_entry(connection, publish=True)


class MstpInfo(_RedisModel):
    """Store MSTP non-port-specific info for troubleshooting
    """
    bridge_name = objects.String(json_key='bridgeName')
    state = objects.Boolean(json_key='state')
    root_port_name = objects.String(json_key='rootPortName')
    root_bridge = objects.String(json_key='rootBridge')
    topo_chg_ports = objects.Json(json_key='topoChangePort')
    time_since_tcn = objects.Long(json_key='timeSinceLastTcn')
    topo_chg_cntr = objects.Long(json_key='toptChangeCntr')
    bridge_id = objects.String(json_key='bridgeId')
    edge_ports = objects.Json(json_key='edgePorts')
    network_ports = objects.Json(json_key='networkPorts')
    disputed_ports = objects.Json(json_key='disputedPorts')
    bpduguard_ports = objects.Json(json_key='bpduguardPorts')
    bpduguard_err_ports = objects.Json(json_key='bpduguardErrPorts')
    ba_inconsistent_ports = objects.Json(json_key='baInconsistentPorts')
    bpdufilter_ports = objects.Json(json_key='bpdufilterPorts')
    is_vlan_filtering = objects.Boolean(json_key='isVlanFiltering')
    ports = objects.Json(json_key='ports')
    app_name = 'mstp'
    api_key_map = {'bridge_name': 'bridgename'}

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'bridge_name']

    @classmethod
    def exclude_fmt(cls):
        return cls.timestamp, cls.active, cls.time_since_tcn

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.bridge_name

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.root_port_name, cls.topo_chg_ports, cls.time_since_tcn,
                cls.topo_chg_cntr, cls.ports, cls.edge_ports, cls.state,
                cls.network_ports, cls.disputed_ports, cls.bpduguard_ports,
                cls.root_bridge, cls.bridge_id, cls.bpduguard_err_ports,
                cls.ba_inconsistent_ports, cls.bpdufilter_ports,
                cls.is_vlan_filtering, cls.active, cls.deleted,
                cls.timestamp)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(MstpInfo, self).save_entry(connection, publish=True)


class OS(_Inventory):
    name = objects.String(json_key='name')
    version = objects.String(json_key='fullVersion')
    version_id = objects.String(json_key='version')

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.timestamp, cls.name, cls.version,
                cls.version_id)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(OS, self).save_entry(connection, publish=True)


class OspfNbr(_RedisModel):
    peer_id = objects.String(json_key='peerId')
    peer_addr = objects.String(json_key='peerAddress')
    state = objects.String(json_key='state')
    ifname = objects.String(json_key='interface')
    is_ipv6 = objects.String(json_key='isIPv6')
    area = objects.String(json_key='area')
    app_name = 'ospfnbr'

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.ifname, cls.peer_id, cls.area

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'ifname', 'peer_id', 'area']

    @classmethod
    def val_fmt(cls):
        return (cls.state, cls.timestamp, cls.ifname, cls.is_ipv6,
                cls.peer_addr, cls.area, cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(OspfNbr, self).save_entry(connection, publish=True)


class OspfIf(_RedisModel):
    ifname = objects.String(json_key='interface')
    is_unnumbered = objects.Boolean(json_key='isUnnumbered')
    if_up = objects.Boolean(json_key='ifUp')
    area = objects.String(json_key='area')
    network_type = objects.String(json_key='networkType')
    nbr_count = objects.Integer(json_key='nbrCount')
    nbr_adj_count = objects.Integer(json_key='nbrAdjCount')
    cost = objects.Integer(json_key='cost')
    is_passive = objects.Boolean(json_key='isPassive')
    mtu = objects.Integer(json_key='mtu')
    dead_time = objects.Integer(json_key='deadTime')
    rexmit_time = objects.Integer(json_key='rexmitTime')
    hello_time = objects.Integer(json_key='helloTime')
    router_id = objects.String(json_key='routerId')
    app_name = 'ospfif'

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.ifname, cls.area

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'ifname', 'area']

    @classmethod
    def val_fmt(cls):
        return (cls.network_type, cls.timestamp, cls.nbr_count, cls.if_up,
                cls.nbr_adj_count, cls.is_unnumbered, cls.is_passive, cls.cost,
                cls.mtu, cls.dead_time, cls.rexmit_time, cls.hello_time,
                cls.router_id, cls.area, cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(OspfIf, self).save_entry(connection, publish=True)


class Neighbor(_RedisModel):
    message_type = objects.String(json_key='messageType')
    ifname = objects.String(json_key='interface')
    ifindex = objects.Integer(json_key='ifIndex')
    ip_address = objects.String(json_key='ipAddress')
    mac_address = objects.MacAddress(json_key='macAddress')
    is_ipv6 = objects.Boolean(json_key='isIPv6')
    vrf = objects.String(json_key='vrf')
    is_remote = objects.Boolean(json_key='isRemote')
    api_filter_args = ['ifname', 'mac_address']
    api_key_map = {'ip_address': 'ipaddress'}
    app_name = 'neighbor'

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return (cls.optional_args_map.get(key, key), val)
        else:
            return ['is_ipv6']

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'ip_address', 'vrf']

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.ifname,
                cls.ip_address, cls.mac_address, cls.is_ipv6, cls.vrf)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.message_type, cls.ifindex, cls.timestamp, cls.is_remote,
                cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(Neighbor, self).save_entry(connection)


class Node(_RedisModel):
    lastboot = objects.Float(json_key='lastBoot')
    sys_uptime = objects.Float(json_key='sysUptime')
    last_reinit = objects.Float(json_key='lastReinit')
    ntp_state = objects.String(json_key='ntpState')
    version = objects.String(json_key='version')
    app_name = 'node'
    post_request_args = ['node_names']

    def __str__(self):
        return self.hostname

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return (cls.optional_args_map.get(key, key), val)
        else:
            return ['filter_rotten', 'filter_dead']

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['path', 'hostname']

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    def get_priority(self):
        return 1

    @classmethod
    def val_fmt(cls):
        return (cls.message_type, cls.lastboot, cls.sys_uptime, cls.last_reinit,
                cls.ntp_state, cls.version, cls.timestamp)

    def save(self, commit=False):
        if self.redis_enabled:
            connection = self._get_connection_object(commit)
            key = Node.obj_to_key(self.__dict__)
            score = Node.obj_to_score(self.__dict__)
            val = Node.obj_to_val(self.__dict__)
            connection.zadd(key, score, val)
            connection.hset(self.__class__.SEPERATOR.join((
                self.__class__._namespace,
                self.__class__.__name__)),
                key, True)
        if self.opta_enabled:
            self.save_grpc()


class Ntp(_RedisModel):
    message_type = objects.String(json_key='messageType')
    current_server = objects.String(json_key='currentServer')
    stratum = objects.Integer(json_key='stratum')
    ntp_sync = objects.String(json_key='ntpSync')
    ntp_app = objects.String(json_key='ntpApp')
    optional_args_map = {'sync_state': 'in_sync'}
    app_name = 'ntp'

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            if val == 'in-sync':
                return (cls.optional_args_map.get(key, key), 'yes')
            elif val == 'out-of-sync':
                return (cls.optional_args_map.get(key, key), 'no')
        else:
            return ['sync_state']

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname']

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.message_type, cls.current_server, cls.stratum,
                cls.ntp_sync, cls.timestamp, cls.ntp_app,
                cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(Ntp, self).save_entry(connection, publish=True)


class Port(_RedisModel):
    ifname = objects.String(json_key='interface')
    identifier = objects.String(json_key='module')
    speed = objects.String(json_key='speed')
    autoneg = objects.String(json_key='autoneg')
    state = objects.String(json_key='state')
    transreceiver = objects.String(json_key='transceiver')
    connector = objects.String(json_key='connector')
    vendor_name = objects.String(json_key='vendor')
    part_number = objects.String(json_key='partNo')
    serial_number = objects.String(json_key='serialNo')
    length = objects.String(json_key='length')
    supported_fec = objects.String(json_key='supportedFEC')
    advertised_fec = objects.String(json_key='advertisedFEC')
    fec = objects.String(json_key='FEC')
    app_name = 'port'

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'ifname']

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.ifname

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.identifier, cls.speed, cls.autoneg, cls.state,
                cls.transreceiver, cls.connector, cls.length, cls.vendor_name,
                cls.part_number, cls.serial_number, cls.fec, cls.supported_fec, cls.advertised_fec,
                cls.active, cls.deleted, cls.timestamp)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(Port, self).save_entry(connection, publish=True)


class ProcNetDevInterface(_RedisModel):
    ifname = objects.String(json_key='interface')
    rx_bytes = objects.Long(json_key='rxBytes')
    rx_drop = objects.Long(json_key='rxDrop')
    rx_errs = objects.Long(json_key='rxErrs')
    rx_frame = objects.Long(json_key='rxFrame')
    rx_multicast = objects.Long(json_key='rxMulticast')
    rx_packets = objects.Long(json_key='rxPackets')
    tx_bytes = objects.Long(json_key='txBytes')
    tx_carrier = objects.Long(json_key='txCarrier')
    tx_colls = objects.Long(json_key='txColls')
    tx_drop = objects.Long(json_key='txDrop')
    tx_errs = objects.Long(json_key='txErrs')
    tx_packets = objects.Long(json_key='txPackets')
    app_name = 'procdevstats'

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'ifname']

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.ifname

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.rx_bytes, cls.rx_drop, cls.rx_errs,
                cls.rx_frame, cls.rx_multicast, cls.rx_packets,
                cls.tx_bytes, cls.tx_carrier, cls.tx_colls,
                cls.tx_drop, cls.tx_errs, cls.tx_packets,
                cls.active, cls.deleted, cls.timestamp)

    def counter_fmt(self):
        return {
            'measurement': 'procnetdev',
            'time': datetime.fromtimestamp(self.timestamp).strftime('%Y-%m-%d %H:%M:%S.%f'),
            'tags': {
                'hostname': self.hostname,
                'interface': self.ifname,
            },
            'fields': {
                'rxBytes': self.rx_bytes,
                'rxDrop': self.rx_drop,
                'rxErrs': self.rx_errs,
                'rxFrame': self.rx_frame,
                'rxMulticast': self.rx_multicast,
                'rxPackets': self.rx_packets,
                'txBytes': self.tx_bytes,
                'txCarrier': self.tx_carrier,
                'txColls': self.tx_colls,
                'txDrop': self.tx_drop,
                'txErrs': self.tx_errs,
                'txPackets': self.tx_packets,
            },
        }

    def counter_retention_policy(self):
        # Delete the downsampled 5m data after one year
        return 'CREATE RETENTION POLICY "one_year" ON "netq" DURATION 52w REPLICATION 1'

    def counter_continuous_query(self):

        # Aggregate the raw 15s data from the previous 5m, this runs ever 5m
        return """
CREATE CONTINUOUS QUERY "cq_5m" ON "netq" BEGIN
  SELECT
    MEAN("rxBytes") AS "mean_rxBytes",
    MEAN("rxDrop") AS "mean_rxDrop",
    MEAN("rxErrs") AS "mean_rxErrs",
    MEAN("rxFrame") AS "mean_rxFrame",
    MEAN("rxMulticast") AS "mean_rxMulticast",
    MEAN("rxPackets") AS "mean_rxPackets",
    MEAN("txBytes") AS "mean_txBytes",
    MEAN("txCarrier") AS "mean_txCarrier",
    MEAN("txColls") AS "mean_txColls",
    MEAN("txDrop") AS "mean_txDrop",
    MEAN("txErrs") AS "mean_txErrs",
    MEAN("txPackets") AS "mean_txPackets"
  INTO "one_year"."downsampled_stats"
  FROM "procnetdev"
  GROUP BY time(5m)
END
"""

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(ProcNetDevInterface, self).save_entry(connection, publish=False)


class PSU(_Sensor):
    message_type = objects.String(json_key='messageType')
    s_name = objects.String(json_key='name')
    s_state = objects.String(json_key='state')
    s_prev_state = objects.String(json_key='prevState')
    s_msg = objects.String(json_key='message')

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.s_name

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.message_type, cls.timestamp, cls.s_state, cls.s_prev_state,
                cls.s_msg, cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(PSU, self).save_entry(connection, publish=True)


class Route(_RedisModel):
    message_type = objects.String(json_key='messageType')
    prefix = objects.String(json_key='prefix')
    protocol = objects.String(json_key='protocol')
    route_type = objects.Integer(json_key='routeType')
    origin = objects.Boolean(json_key='origin')
    nexthops = objects.Json(json_key='nextHops')
    src = objects.String(json_key='src')
    is_ipv6 = objects.Boolean(json_key='isIPv6')
    rt_table_id = objects.Integer(json_key='routeTableId')
    vrf = objects.String(json_key='vrf')
    priority = objects.Integer(json_key='priority')
    api_key_map = {'prefix': 'ipaddress'}
    optional_args_map = {'origin': 'is_origin'}
    app_name = 'route'

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return (cls.optional_args_map.get(key, key), val)
        else:
            return ['origin', 'is_ipv6']

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'lpm', 'prefix', 'vrf']

    @classmethod
    def key_fmt(cls, key_type=None):
        # The exact sequence of key fields here is assumed in the lpm LUA
        # script. Do not change one without changing the other. We need both
        # VRF and table in key to handle redistribute neighbor table. Also names
        # are constant across nodes, but table ids may vary from box to box. So,
        # save the routes with the VRF name as well as table id.
        return (cls.__name__, cls.hostname, cls.prefix, cls.route_type,
                cls.rt_table_id, cls.is_ipv6, cls.origin, cls.protocol, cls.vrf)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.message_type, cls.nexthops, cls.src, cls.timestamp, cls.active,
                cls.deleted, cls.priority)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(Route, self).save_entry(connection)


class Services(_RedisModel):
    name = objects.String(json_key='service')
    is_enabled = objects.Boolean(json_key='isEnabled')
    is_active = objects.Boolean(json_key='isActive')
    is_monitored = objects.Boolean(json_key='isMonitored')
    status = objects.Integer(json_key='status')
    vrf = objects.String(json_key='vrf')
    pid = objects.String(json_key='pid')
    start_time = objects.Float(json_key='startTime')
    optional_args_map = {'is_active': 'active', 'is_monitored': 'monitored', 'stat_string': 'status'}
    api_key_map = {'name': 'type'}
    app_name = 'services'

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            if val == 'active' or val == 'monitored':
                return (cls.optional_args_map.get(key, key), 'yes')
            elif val == 'ok':
                return (cls.optional_args_map.get(key, key), ServiceStatus.OK)
            elif val == 'warning':
                return (cls.optional_args_map.get(key, key), ServiceStatus.WARNING)
            elif val == 'error':
                return (cls.optional_args_map.get(key, key), ServiceStatus.ERROR)
            elif val == 'fail':
                return (cls.optional_args_map.get(key, key), ServiceStatus.FAIL)
            else:
                return (cls.optional_args_map.get(key, key), val)
        else:
            return ['is_active', 'is_monitored', 'stat_string']

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'name', 'vrf']

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.name, cls.vrf)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.is_enabled, cls.is_active, cls.status, cls.is_monitored,
                cls.start_time, cls.pid,
                cls.timestamp, cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(Services, self).save_entry(connection, publish=True)


class Temp(_Sensor):
    message_type = objects.String(json_key='messageType')
    s_name = objects.String(json_key='name')
    s_desc = objects.String(json_key='description')
    s_input = objects.Float(json_key='temp')
    s_crit = objects.Integer(json_key='critical')
    s_lcrit = objects.Integer(json_key='lcritical')
    s_max = objects.Integer(json_key='max')
    s_min = objects.Integer(json_key='min')
    s_state = objects.String(json_key='state')
    s_prev_state = objects.String(json_key='prevState')
    s_msg = objects.String(json_key='message')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.s_name,
                cls.s_desc)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.message_type, cls.timestamp, cls.s_state, cls.s_prev_state,
                cls.s_input, cls.active, cls.deleted,
                cls.s_msg, cls.s_crit, cls.s_max, cls.s_min, cls.s_lcrit)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(Temp, self).save_entry(connection, publish=True)


class VxlanRemoteDest(_RedisModel):
    vni = objects.Integer(json_key='vni')
    rdst = objects.String(json_key='remoteDest')
    app_name = 'vxlan'

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.vni, cls.rdst

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'vni', 'rdst']

    @classmethod
    def val_fmt(cls):
        return (cls.vni, cls.rdst, cls.active, cls.deleted, cls.timestamp)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(VxlanRemoteDest, self).save_entry(connection, publish=True)

# DATA FROM APPS ONLY #


class Vlan(_RedisModel):
    vlan = objects.String(json_key='vlan')
    ports = objects.String(json_key='ports')
    svi = objects.String(json_key='svi')
    app_name = 'vlan'

    @classmethod
    def key_fmt(cls, key_type=None):
        return []

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return []

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'vlan']


class Macs(_RedisModel):
    origin = objects.Boolean(json_key='origin')
    mac_address = objects.MacAddress(json_key='macAddress')
    vlan = objects.Integer(json_key='vlan')
    egress_port = objects.String(json_key='egressPort')
    remote = objects.Boolean(json_key='remote')
    app_name = 'macs'
    api_key_map = {'mac_address': 'macaddress'}

    @classmethod
    def key_fmt(cls, key_type=None):
        return []

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return []

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'mac_address', 'vlan', 'origin']


class Interface(_RedisModel):
    details = objects.String(json_key='details')
    ifname = objects.String(json_key='interface')
    state = objects.String(json_key='state')
    type = objects.String(json_key='vlan')
    vrf = objects.String(json_key='vrf')
    app_name = 'interface'
    api_filter_args = ['ifname', 'kind', 'state']

    @classmethod
    def key_fmt(cls, key_type=None):
        return []

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return []

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'ifname', 'kind', 'state']


class Notification(_RedisModel):
    app_name = 'notifier'

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return (cls.optional_args_map.get(key, key), val)
        else:
            return ['filter', 'enable_slack', 'enable_pagerduty', 'proxy_url']

    @classmethod
    def key_fmt(cls, key_type=None):
        return []

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return []

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['path']


class Event(_RedisModel):
    severity = objects.String(json_key='severity')
    message_id = objects.Integer(json_key='messageId')
    filter_message = objects.String(json_key='filterMessage')
    message = objects.String(json_key='message')
    app_name = 'events'

    @classmethod
    def key_fmt(cls, key_type=None):
        return []

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return []

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return []

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return (cls.optional_args_map.get(key, key), val)
        else:
            return ['hostnames', 'severity', 'message_types']


class Filter(Notification):
    delete_request_args = ['filter']
    post_request_args = ['filter_name', 'channel_names', 'rule_names', 'terminate_on_match', 'severity', 'order_type',
                         'order_filter_name']


class Rule(Notification):
    delete_request_args = ['rule']
    post_request_args = ['rule_name', 'rule_json']


class Channel(Notification):
    delete_request_args = ['channel']
    post_request_args = ['channel_type', 'channel_name', 'severity', 'channel_info_json']


class Search(_RedisModel):
    hostname = objects.String(json_key='hostname')
    item_type = objects.String(json_key='itemType')
    ip_address = objects.String(json_key='ipAddress')
    app_name = 'search'

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return (cls.optional_args_map.get(key, key), val)
        else:
            return ['prefix']

    @classmethod
    def key_fmt(cls, key_type=None):
        return []

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return []

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['path']


class Validation(_RedisModel):
    app_name = 'check'
    validation_type = objects.Json(json_key='validationType')
    post_request_args = ['validation_type', 'cadence_min', 'alert_on_failure',
                         'start_time', 'validation_name', 'action']

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return (cls.optional_args_map.get(key, key), val)
        else:
            return ['type', 'by', 'proto', 'time', 'duration']

    @classmethod
    def key_fmt(cls, key_type=None):
        return []

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return []

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['path']


class Trace(_RedisModel):
    app_name = 'trace'
    post_request_args = ['data']

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return (cls.optional_args_map.get(key, key), val)
        else:
            return ['type', 'time', 'duration', 'bucket', 'by']

    @classmethod
    def key_fmt(cls, key_type=None):
        return []

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return []

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['path']


class Label(_RedisModel):
    app_name = 'label'
    put_request_args = ['labelname', 'labeltype', 'labelmap']

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return (cls.optional_args_map.get(key, key), val)
        else:
            return []

    @classmethod
    def key_fmt(cls, key_type=None):
        return []

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return []

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['labelname']
